# -*- coding: utf-8 -*-
from .input import create_input, write_input  # noqa

try:
    # only import the chemiscope.show function if we have ipywidgets installed.
    import ipywidgets  # noqa
    from .jupyter import show  # noqa
except ImportError:
    pass

__version__ = "0.3.3"



# -*- coding: utf-8 -*-
import numpy as np
import warnings

from collections import Counter

try:
    import ase

    HAVE_ASE = True
except ImportError:
    HAVE_ASE = False


def frames_to_json(frames):
    """
    Convert the given ``frames`` to the JSON structure used by chemiscope.

    This function is a shim calling specialized implementations for all the
    supported frame types. Currently only `ase.Atoms` frames are supported.

    :param frames: iterable over structures (typically a list of frames)
    """
    frames_list = list(frames)

    if HAVE_ASE and isinstance(frames_list[0], ase.Atoms):
        return [_ase_to_json(frame) for frame in frames_list]
    elif HAVE_ASE and isinstance(frames_list[0], ase.Atom):
        # deal with the user passing a single frame
        return frames_to_json([frames])
    else:
        raise Exception(f"unknown frame type: '{frames_list[0].__class__.__name__}'")


def atom_properties(frames, composition):
    """
    Extract "atom" properties from the given ``frames``, and give them as a
    dictionary compatible with :py:func:`create_input`.

    This function is a shim calling specialized implementations for all the
    supported frame types. Currently only `ase.Atoms` frames are supported.

    :param frames: iterable over structures (typically a list of frames)
    """
    frames_list = list(frames)

    if HAVE_ASE and isinstance(frames_list[0], ase.Atoms):
        return _ase_atom_properties(frames_list, composition)
    elif HAVE_ASE and isinstance(frames_list[0], ase.Atom):
        # deal with the user passing a single frame
        return atom_properties([frames], composition)
    else:
        raise Exception(f"unknown frame type: '{frames_list[0].__class__.__name__}'")


def structure_properties(frames, composition):
    """
    Extract "structure" properties from the given ``frames``, and give them as a
    dictionary compatible with :py:func:`create_input`.

    This function is a shim calling specialized implementations for all the
    supported frame types. Currently only `ase.Atoms` frames are supported.

    :param frames: iterable over structures (typically a list of frames)
    """
    frames_list = list(frames)

    if HAVE_ASE and isinstance(frames_list[0], ase.Atoms):
        return _ase_structure_properties(frames_list, composition)
    elif HAVE_ASE and isinstance(frames_list[0], ase.Atom):
        # deal with the user passing a single frame
        return structure_properties([frames], composition)
    else:
        raise Exception(f"unknown frame type: '{frames_list[0].__class__.__name__}'")


def _add_structure_chemical_composition(frames):
    """Adds information about the chemical composition and count for every chemical species
    to the info dictionary for every frame as a structure property"""
    all_elements = set()
    for frame in frames:
        all_elements.update(frame.symbols)
    all_elements = set(all_elements)
    for frame in frames:
        frame.info["composition"] = str(frame.symbols)
        dict_composition = dict(Counter(frame.symbols))
        for element in all_elements:
            if element in dict_composition:
                frame.info[f"n_{element}"] = dict_composition[element]
            else:
                frame.info[f"n_{element}"] = 0
    return frames


def _add_atom_chemical_composition(frames):
    """Adds information about the atoms chemical species to the frame.arrays
    for every frame as an atom property"""
    for frame in frames:
        dict_elements = {"element": list(frame.symbols)}
        frame.arrays.update(dict_elements)
    return frames


def _ase_to_json(frame):
    """Implementation of frame_to_json for ase.Atoms"""
    data = {}
    data["size"] = len(frame)
    data["names"] = list(frame.symbols)
    data["x"] = [float(value) for value in frame.positions[:, 0]]
    data["y"] = [float(value) for value in frame.positions[:, 1]]
    data["z"] = [float(value) for value in frame.positions[:, 2]]

    if (frame.cell.lengths() != [0.0, 0.0, 0.0]).all():
        data["cell"] = list(np.concatenate(frame.cell))

    return data


def _ase_atom_properties(frames, composition):
    """Implementation of atom_properties for ase.Atoms"""
    IGNORED_ASE_ARRAYS = ["positions", "numbers"]
    # extract the set of common properties between all frames
    all_names = set()
    extra = set()

    if composition:
        _add_atom_chemical_composition(frames)

    for name in frames[0].arrays.keys():
        if name in IGNORED_ASE_ARRAYS:
            continue
        all_names.add(name)

    for frame in frames[1:]:
        for name in frame.arrays.keys():
            if name in IGNORED_ASE_ARRAYS:
                continue

            if name not in all_names:
                extra.add(name)

        remove = []
        for name in all_names:
            if name not in frame.arrays.keys():
                remove.append(name)

        for name in remove:
            all_names.remove(name)
            extra.add(name)

    if len(extra) != 0:
        warnings.warn(
            "the following atomic properties properties are only defined "
            + f"for a subset of frames: {list(sorted(extra))}; they will be ignored"
        )

    # create property in the format expected by create_input
    properties = {
        name: {"target": "atom", "values": value}
        for name, value in frames[0].arrays.items()
        if name in all_names
    }

    for frame in frames[1:]:
        for name, value in frame.arrays.items():
            if name not in all_names:
                continue
            properties[name]["values"] = np.concatenate(
                [properties[name]["values"], value]
            )

    _remove_invalid_properties(properties, "ASE")
    return properties


def _ase_structure_properties(frames, composition):
    """Implementation of structure_properties for ase.Atoms"""
    # extract the set of common properties between all frames
    all_names = set()
    extra = set()

    if composition:
        _add_structure_chemical_composition(frames)

    for name in frames[0].info.keys():
        all_names.add(name)

    for frame in frames[1:]:
        for name in frame.info.keys():
            if name not in all_names:
                extra.add(name)

        remove = []
        for name in all_names:
            if name not in frame.info.keys():
                remove.append(name)

        for name in remove:
            all_names.remove(name)
            extra.add(name)

    if len(extra) != 0:
        warnings.warn(
            "the following structure properties properties are only defined "
            + f"for a subset of frames: {list(sorted(extra))}; they will be ignored"
        )

    # create property in the format expected by create_input
    properties = {name: {"target": "structure", "values": []} for name in all_names}

    for frame in frames:
        for name, value in frame.info.items():
            if name in all_names:
                properties[name]["values"].append(value)

    _remove_invalid_properties(properties, "ASE")

    return properties


def _remove_invalid_properties(properties, origin):
    """
    Remove invalid properties from the ``properties`` dictionary. ``origin`` is
    used in error messages as the property origin
    """
    to_remove = []
    for name, property in properties.items():
        for value in property["values"]:
            if not _is_convertible_to_property(value):
                warnings.warn(
                    f"value '{value}' of type '{type(value)}' for the '{name}' "
                    + f"property from {origin} is not convertible to float or "
                    + "string, this property will be ignored."
                )
                to_remove.append(name)
                break

    for name in to_remove:
        del properties[name]


def _is_convertible_to_property(value):
    """
    Check whether a value is convertible to a chemiscope property, i.e. if it is
    a string or something convertible to float.
    """
    if isinstance(value, (bytes, str)):
        # string types
        return True
    else:
        # everything convertible to float
        try:
            float(value)
            return True
        except Exception:
            return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .input import write_input


def main():
    """
    Command-line utility to generate an input for chemiscope — the interactive
    structure-property explorer. Parses an input file containing atomic
    structures using the ASE I/O module, and converts it into a JSON file that
    can be loaded in chemiscope. Frame and environment properties must be
    written in the same file containing atomic structures: we recommend the
    extended xyz format, which is flexible and simple. In all cases, this
    utility will simply write to the JSON file anything that is readable by
    ASE.
    """

    import argparse

    try:
        # command-line execution. requires ASE IO module
        import ase.io as ase_io
    except ImportError:
        raise ImportError(
            "chemiscope_input needs ASE modules to parse structure inputs"
        )

    # Tweak the autogenerated help output to look nicer
    def formatter(prog):
        return argparse.HelpFormatter(prog, max_help_position=22)

    parser = argparse.ArgumentParser(
        description=main.__doc__, formatter_class=formatter
    )
    parser.add_argument(
        "input", type=str, help="input file containing the structures and properties"
    )
    parser.add_argument(
        "-o", "--output", type=str, help="chemiscope output file in JSON format"
    )
    parser.add_argument(
        "-c",
        "--cutoff",
        type=float,
        help="generate atom-centred environments with the given cutoff",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--only-atoms",
        action="store_true",
        help="only use per-atom properties from the input file",
    )
    group.add_argument(
        "--only-structures",
        action="store_true",
        help="only use per-structure properties from the input file",
    )
    parser.add_argument("--name", default="", type=str, help="name of the dataset")
    parser.add_argument(
        "--description", default="", type=str, help="description of the dataset"
    )
    parser.add_argument(
        "--authors", nargs="*", type=str, default=[], help="list of dataset authors"
    )
    parser.add_argument(
        "--references",
        nargs="*",
        type=str,
        default=[],
        help="list of references for the dataset",
    )
    args = parser.parse_args()

    if args.only_atoms and args.cutoff is None:
        raise Exception("--only-atoms requires to give --cutoff")
    if args.only_structures and args.cutoff is not None:
        raise Exception("--only-structure can not be given with --cutoff")

    # read file with ASE and remove extraneous properties
    frames = ase_io.read(args.input, ":")
    if args.only_structures:
        for frame in frames:
            for key in list(frame.arrays.keys()):
                if key not in ["positions", "numbers"]:
                    del frame.arrays[key]
    elif args.only_atoms:
        for frame in frames:
            frame.info = {}

    # determine output file name automatically if missing
    output = args.output or args.input + "_chemiscope.json.gz"

    write_input(
        filename=output,
        frames=frames,
        meta={
            "name": args.name,
            "description": args.description,
            "authors": args.authors,
            "references": args.references,
        },
        extra=None,
        cutoff=args.cutoff,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from setuptools import setup
from distutils.command import sdist, install
from wheel import bdist_wheel

import subprocess
import os
import sys

# customize dist directory if the user did not already request one, and if we
# are running setup.py directly (i.e. not from pip)
if not sys.argv[0].endswith("setup.py"):
    change_dist_dir = False

change_dist_dir = True
for arg in sys.argv:
    if "--dist-dir" in arg:
        change_dist_dir = False


def run_npm_build():
    """Build the JavaScript code required by the jupyter widget using npm"""
    root = os.path.dirname(os.path.realpath(__file__))

    # This file does not exists when building a wheel (for pip installation)
    # from a sdist. This is fine since the sdist already contains the required
    # js files.
    if os.path.exists(os.path.join(root, "package.json")):
        subprocess.run("npm ci", check=True, shell=True)
        subprocess.run("npm run build:nbextension", check=True, shell=True)


class sdist_with_npm(sdist.sdist):
    def run(self):
        if change_dist_dir:
            self.dist_dir = os.path.join("dist", "python")
        run_npm_build()
        super().run()


class bdist_wheel_with_npm(bdist_wheel.bdist_wheel):
    def run(self):
        run_npm_build()
        if change_dist_dir:
            self.dist_dir = os.path.join("dist", "python")
        super().run()


class install_with_npm(install.install):
    def run(self):
        run_npm_build()
        super().run()


if __name__ == "__main__":
    setup(
        cmdclass={
            # we need to override multiple classes here to cover all possible
            # use cases. `sdist` is used to generate the source upload to PyPI,
            # `bdist_wheel` is used both to generate the wheel from the sdist
            # (pip install from PyPI) and the wheel from the full repository
            # (pip install with local path); and `install` adds support for
            # `python setup.py install`. Notablely missing here is `python
            # setup.py develop`/`pip install -e <local/path>`.
            "sdist": sdist_with_npm,
            "bdist_wheel": bdist_wheel_with_npm,
            "install": install_with_npm,
        },
        data_files=[
            # this is what `jupyter nbextension install --sys-prefix` does
            (
                "share/jupyter/nbextensions/chemiscope-widget",
                [
                    "python/nbextension/extension.js",
                    "python/nbextension/build/chemiscope-widget.min.js",
                ],
            ),
            # this is what `jupyter nbextension enable --sys-prefix` does
            (
                "etc/jupyter/nbconfig/notebook.d",
                [
                    "python/nbextension/chemiscope-widget.json",
                ],
            ),
        ],
    )
